@@ -1,6 +1,7 @@
 __author__ = "ardevelop"
 
 import tornado.ioloop
+import time
 
 import psycopg2
 import psycopg2.extras
@@ -9,6 +10,12 @@ import psycopg2.extensions
 from psycopg2.extras import NamedTupleCursor, DictCursor, RealDictCursor
 
 
+def success_callback(callback):
+    def _callback(cursor):
+        callback(bool(cursor))
+    return _callback
+
+
 def fetchone_callback(callback):
     def _callback(cursor):
         callback(cursor.fetchone())
@@ -56,21 +63,34 @@ class PostgreAsyncConnection():
             self._ioloop.add_callback(self._connect_callback, (conn,))
 
         if self._init_chain:
-            self._init_chain.pop(0)(self, _callback)
+            try:
+                self._init_chain.pop(0)(self, _callback)
+            except Exception, ex:
+                try:
+                    self._pool._connect_error(self)
+                except AttributeError:
+                    pass
+
+                raise ex
 
     def _io_callback(self, fd, events):
         try:
             state = self._conn.poll()
-        except psycopg2.OperationalError, ex:
-            self._ioloop.remove_handler(fd)
-            try:
-                self._pool._disconnect(self)
-            except AttributeError:
-                pass
-            raise ex
         except (psycopg2.Warning, psycopg2.Error), ex:
             self._ioloop.remove_handler(fd)
-            raise ex
+            try:
+                if ex.message.endswith("connection failed"):
+                    self._pool._connect_error(self)
+                    raise ex
+                else:
+                    self._pool._disconnect(self)
+            except AttributeError:
+                pass
+
+            if self._pool is not None:
+                self._callback(ex)
+            else:
+                raise ex
         else:
             if psycopg2.extensions.POLL_OK == state:
                 self._ioloop.remove_handler(fd)
@@ -114,13 +134,14 @@ class PostgreAsyncConnection():
 
 class PostgreAsyncClient(object):
     def __init__(self, dsn=None, database=None, username=None, password=None, host=None, port=None,
-                 min_pool_size=1, max_pool_size=1, ioloop=None, connection_factory=None, cursor_factory=None,
+                 min_pool_size=0, max_pool_size=1, ioloop=None, connection_factory=None, cursor_factory=None,
                  hstore=False, composite_types=None, init=None, **kwargs):
         self._ioloop = ioloop or tornado.ioloop.IOLoop.instance()
         self._pool = set()
         self._wait_connections = set()
         self._min_pool_size = min_pool_size
         self._max_pool_size = max_pool_size
+        self._last_connect_error = None
 
         init_chain = []
 
@@ -145,7 +166,7 @@ class PostgreAsyncClient(object):
             else:
                 init_chain.append(init)
 
-        init_chain.append(self._initialized_callback)
+        init_chain.append(self._connect_success)
 
         def connect():
             conn = PostgreAsyncConnection(dsn, database, username, password, host, port, self, connection_factory,
@@ -196,30 +217,49 @@ class PostgreAsyncClient(object):
 
         return _register_composite_task
 
-    def _initialized_callback(self, conn, callback):
+    def _connect_success(self, conn, callback):
         self._pool.add(conn)
         self._wait_connections.remove(conn)
 
+    def _connect_error(self, conn):
+        self._last_connect_error = time.time()
+        conn.close()
+        self._wait_connections.remove(conn)
+
     def _acquire_connection(self):
         for conn in self._pool:
+            if conn.closed():
+                self._disconnect(conn)
+                continue
+
             if not conn.busy():
                 return conn
 
         if len(self._wait_connections) == 0 and len(self._pool) < self._max_pool_size:
+            if self._last_connect_error and time.time() - self._last_connect_error < 5:
+                raise Exception("Connect error retry in 5 seconds")
+
             self._connect()
 
     def _disconnect(self, conn):
-        self._pool.remove(conn)
         conn.close()
+
         try:
-            self._wait_connections.remove(conn)
+            self._pool.remove(conn)
         except KeyError:
             pass
 
     def execute(self, statement, args=(), cur_factory=None, callback=None):
         conn = self._acquire_connection()
+
+        def _callback(response):
+            if isinstance(response, (psycopg2.Warning, psycopg2.Error)):
+                self._ioloop.add_callback(self.execute, statement, args, cur_factory, callback)
+            else:
+                callback(response)
+
         if conn:
-            conn.execute(statement, args, cur_factory, callback)
+            conn.execute(statement, args, cur_factory, _callback)
         else:
             self._ioloop.add_callback(self.execute, statement, args, cur_factory, callback)
 
